# vim:  tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013 Embrane, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author:  Ivar Lazzaro, Embrane, Inc.

import sys

import mock
from oslo.config import cfg

from neutron.plugins.embrane.common import config  # noqa
from neutron.tests import base

# Need to mock heleosapi.
sys.modules["heleosapi"] = mock.Mock()


class ConfigurationTest(base.BaseTestCase):

    def test_defaults(self):
        self.assertEqual('admin', cfg.CONF.heleos.admin_username)
        self.assertEqual('default', cfg.CONF.heleos.resource_pool_id)
        self.assertTrue(cfg.CONF.heleos.async_requests)
