# Copyright (C) 2014 VA Linux Systems Japan K.K.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# @author: Fumihiko Kakuma, VA Linux Systems Japan K.K.

from oslo.config import cfg

from neutron.plugins.ofagent.common import config  # noqa
from neutron.tests import base


class ConfigurationTest(base.BaseTestCase):
    """Configuration file Tests."""
    def test_ml2_defaults(self):
        self.assertEqual(60, cfg.CONF.AGENT.get_datapath_retry_times)
